import logging
from typing import Type, Dict, Any, Optional, Hashable, TYPE_CHECKING, Generic, Literal

from slixmpp import Message, Presence, JID
from slixmpp.exceptions import XMPPError

from ..db import GatewayUser, user_store
from ..util import BiDict, ABCSubclassableOnceAtMost
from .contact import LegacyContactType, LegacyRosterType, LegacyRoster

if TYPE_CHECKING:
    from ..gateway import BaseGateway


class BaseSession(
    Generic[LegacyContactType, LegacyRosterType], metaclass=ABCSubclassableOnceAtMost
):
    """
    Represents a gateway user logged in to the network and performing actions.

    Must be overridden for a functional slidge plugin
    """

    store_sent = True
    """
    Keep track of sent messages. Useful to later update the messages' status, e.g.,
    with a read mark from the recipient
    """

    xmpp: "BaseGateway"

    def __init__(self, user: GatewayUser):
        self._roster_cls: Type[
            LegacyRosterType
        ] = LegacyRoster.get_self_or_unique_subclass()

        self.user = user
        if self.store_sent:
            self.sent: BiDict = BiDict()  # TODO: set a max size for this
        self.logged = False

        self.contacts: LegacyRosterType = self._roster_cls(self)
        self.post_init()

    @staticmethod
    def legacy_msg_id_to_xmpp_msg_id(legacy_msg_id: Any):
        return str(legacy_msg_id)

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> Any:
        return i

    def post_init(self):
        """
        Add useful attributes for your session here, if necessary
        """
        pass

    @classmethod
    def _from_user_or_none(cls, user):
        if user is None:
            raise XMPPError(
                text="User not found", condition="subscription-required", etype="auth"
            )

        session = _sessions.get(user)
        if session is None:
            _sessions[user] = session = cls(user)
        return session

    @classmethod
    def from_stanza(cls, s) -> "BaseSession":
        """
        Get a user's :class:`LegacySession` using the "from" field of a stanza

        Ensure that we only have a single session instance per user

        :param s:
        :return:
        """
        return cls._from_user_or_none(user_store.get_by_stanza(s))

    @classmethod
    def from_jid(cls, jid: JID):
        return cls._from_user_or_none(user_store.get_by_jid(jid))

    @classmethod
    async def kill_by_jid(cls, jid: JID):
        log.debug("Killing session of %s", jid)
        for user, session in _sessions.items():
            if user.jid == jid.bare:
                break
        else:
            log.debug("Did not find a session for %s", jid)
            return
        for c in session.contacts:
            c.unsubscribe()
        await session.logout(None)
        await cls.xmpp.unregister(user)
        del _sessions[user]
        del user
        del session

    async def login(self, p: Presence):
        """
        Login the gateway user to the legacy network.

        Triggered when the gateway receives an online presence from the user, so the legacy client
        should keep a list of logged-in users to avoid useless calls to the login process.

        :param p:
        """
        raise NotImplementedError

    async def logout(self, p: Optional[Presence]):
        """
        Logout the gateway user from the legacy network.

        Called when the gateway receives an offline presence from the user.
        Just override this and ``pass`` to implement a bouncer-like ("always connected") functionality.

        :param p:
        """
        raise NotImplementedError

    async def send_from_msg(self, m: Message):
        if m["replace"][
            "id"
        ]:  # ignore last message correction (handled by a specific method)
            return
        url = m["oob"]["url"]
        text = m["body"]
        if url:
            legacy_msg_id = await self.send_file(url, self.contacts.by_stanza(m))
        elif text:
            legacy_msg_id = await self.send_text(text, self.contacts.by_stanza(m))
        else:
            log.debug("Ignoring %s", m)
            return
        self.sent[legacy_msg_id] = m.get_id()

    async def active_from_msg(self, m: Message):
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.active(self.contacts.by_stanza(m))

    async def inactive_from_msg(self, m: Message):
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.inactive(self.contacts.by_stanza(m))

    async def composing_from_msg(self, m: Message):
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.composing(self.contacts.by_stanza(m))

    async def paused_from_msg(self, m: Message):
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.paused(self.contacts.by_stanza(m))

    async def displayed_from_msg(self, m: Message):
        displayed_msg_id = m["displayed"]["id"]
        try:
            legacy_msg_id = self.xmpp_msg_id_to_legacy_msg_id(displayed_msg_id)
        except NotImplementedError:
            log.debug("Couldn't convert xmpp msg ID to legacy ID, ignoring read mark")
            return

        await self.displayed(legacy_msg_id, self.contacts.by_stanza(m))

    async def correct_from_msg(self, m: Message):
        xmpp_id = m["replace"]["id"]
        legacy_id = self.sent.inverse.get(xmpp_id)
        if legacy_id is None:
            log.debug("Did not find legacy ID to correct")
            await self.send_text(m["body"], self.contacts.by_stanza(m))
        else:
            await self.correct(m["body"], legacy_id, self.contacts.by_stanza(m))

    async def send_text(self, t: str, c: LegacyContactType) -> Optional[Hashable]:
        """
        The user wants to send a text message from xmpp to the legacy network

        :param t: Content of the message
        :param c: Recipient of the message
        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    async def send_file(self, u: str, c: LegacyContactType) -> Optional[Hashable]:
        """
        The user has sent a file using HTTP Upload

        :param u: URL of the file
        :param c: Recipient of the file
        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    async def active(self, c: LegacyContactType):
        """
        The use sens an 'active' chat state to the legacy network

        :param c: Recipient of the active chat state
        """
        raise NotImplementedError

    async def inactive(self, c: LegacyContactType):
        """
        The user sends an 'inactive' chat state to the legacy network

        :param c:
        """
        raise NotImplementedError

    async def composing(self, c: LegacyContactType):
        """
        The user is typing

        :param c:
        """
        raise NotImplementedError

    async def paused(self, c: LegacyContactType):
        """
        The user paused typing

        :param c:
        """
        raise NotImplementedError

    async def displayed(self, legacy_msg_id: Any, c: LegacyContactType):
        """
        The user has read a message

        This is only possible if a valid ``legacy_msg_id`` was passed when transmitting a message
        from a contact to the user.

        :param legacy_msg_id: Identifier of the message, passed to :meth:`slidge.LegacyContact.send_text`
            or :meth:`slidge.LegacyContact.send_file`
        :param c:
        :return:
        """
        raise NotImplementedError

    async def correct(self, text: str, legacy_msg_id: Any, c: LegacyContactType):
        """
        The user corrected a message using :xep:`308`

        :param text:
        :param legacy_msg_id:
        :param c:
        :return:
        """
        raise NotImplementedError

    async def search(self, form_values: Dict[str, str]):
        raise NotImplementedError

    def send_gateway_status(
        self,
        status: Optional[str] = None,
        show=Optional[Literal["away", "chat", "dnd", "xa"]],
        **kwargs
    ):
        """
        Send a presence from the gateway to the user.

        Can be used to indicate the user session status, ie "SMS code required", "connected", …

        :param status: A status message
        :param show: Presence stanza 'show' element. I suggest using "dnd" to show
            that the gateway is not fully functional
        """
        self.xmpp.send_presence(
            pto=self.user.bare_jid, pstatus=status, pshow=show, **kwargs
        )

    def send_gateway_message(self, text, **msg_kwargs):
        self.xmpp.send_message(mto=self.user.jid, mbody=text, **msg_kwargs)

    async def input(self, text: str, **msg_kwargs):
        return await self.xmpp.input(self.user, text, **msg_kwargs)

    async def send_qr(self, text: str):
        await self.xmpp.send_qr(text, mto=self.user.jid)


_sessions: Dict[GatewayUser, BaseSession] = {}
log = logging.getLogger(__name__)



from slidge import *

REGISTRATION_INSTRUCTIONS = (
    "Fill the form to use your XMPP account as a signal client. "
    "More information at https://slidge.readthedocs.io/en/latest/user/signal.html"
)
REGISTRATION_FIELDS = [
    FormField(var="phone", label="Phone number (ex: +123456789)", required=True),
    FormField(
        var="device",
        type="list-single",
        label="What do you want to do?",
        options=[
            {"label": "Create a new signal account", "value": "primary"},
            {"label": "Link to an existing signal account", "value": "secondary"},
        ],
        required=True,
    ),
    FormField(
        var="name",
        label="Your name (only used if you chose to create a new signal account; "
        "doesn't have to be your real name)",
    ),
    FormField(
        var="device_name",
        label="Device name (only used if you chose to link to an existing signal account)",
        value="slidge",
    ),
]

CAPTCHA_REQUIRED = (
    "Signal requires you to complete a captcha to register a new account. "
    "Please follow the instructions at https://signald.org/articles/captcha/#getting-a-token and "
    "reply to this message with the token (signalcaptcha://XXXXXXXXX)."
)

NAME_REQUIRED = (
    "If you want to register a new signal account, you must enter a name. "
    "Nothing forces you to use your real name."
)

LINK_TIMEOUT = (
    "You took too much time… "
    "Reply to this message with 'link' once you're ready or 'cancel' "
    "to remove your gateway registration"
)

LINK_SUCCESS = (
    "It looks like everything's all set up. You should now send and "
    "receive signal messages via XMPP."
)

REGISTER_SUCCESS = "This XMPP bridge is now your 'primary' signal device. Congrats!"



import tempfile
import types
from pathlib import Path
from typing import Union

from slixmpp.test import SlixTest, TestTransport

from slidge import user_store, BaseGateway, BaseSession, LegacyRoster, LegacyContact


class SlidgeTest(SlixTest):
    plugin: Union[types.ModuleType, dict]

    class Config:
        jid = "aim.shakespeare.lit"
        secret = "test"
        server = "shakespeare.lit"
        port = 5222
        upload_service = "upload.test"
        home_dir = Path(tempfile.mkdtemp())
        user_jid_validator = ".*@shakespeare.lit"
        admins: list[str] = []

    @classmethod
    def setUpClass(cls):
        user_store.set_file(Path(tempfile.mkdtemp()) / "test.db")

    def setUp(self):
        BaseGateway._subclass = find_subclass(self.plugin, BaseGateway)
        BaseSession._subclass = find_subclass(self.plugin, BaseSession)
        LegacyRoster._subclass = find_subclass(self.plugin, LegacyRoster, base_ok=True)
        LegacyContact._subclass = find_subclass(
            self.plugin, LegacyContact, base_ok=True
        )

        self.xmpp = BaseGateway.get_self_or_unique_subclass()(self.Config)

        self.xmpp._always_send_everything = True

        self.xmpp.connection_made(TestTransport(self.xmpp))
        self.xmpp.session_bind_event.set()
        # Remove unique ID prefix to make it easier to test
        self.xmpp._id_prefix = ""
        self.xmpp.default_lang = None
        self.xmpp.peer_default_lang = None

        def new_id():
            self.xmpp._id += 1
            return str(self.xmpp._id)

        self.xmpp._id = 0
        self.xmpp.new_id = new_id

        # Must have the stream header ready for xmpp.process() to work.
        header = self.xmpp.stream_header

        self.xmpp.data_received(header)
        self.wait_for_send_queue()

        self.xmpp.socket.next_sent()
        self.xmpp.socket.next_sent()

        # Some plugins require messages to have ID values. Set
        # this to True in tests related to those plugins.
        self.xmpp.use_message_ids = False
        self.xmpp.use_presence_ids = False

    @classmethod
    def tearDownClass(cls):
        BaseSession.reset_subclass()
        BaseGateway.reset_subclass()
        LegacyRoster.reset_subclass()
        LegacyContact.reset_subclass()
        user_store._users = None

    def next_sent(self):
        self.wait_for_send_queue()
        sent = self.xmpp.socket.next_sent(timeout=1)
        if sent is None:
            return None
        xml = self.parse_xml(sent)
        self.fix_namespaces(xml, "jabber:component:accept")
        sent = self.xmpp._build_stanza(xml, "jabber:component:accept")
        return sent


def find_subclass(o, parent, base_ok=False):
    try:
        vals = vars(o).values()
    except TypeError:
        vals = o.values()
    for x in vals:
        try:
            if issubclass(x, parent) and x is not parent:
                return x
        except TypeError:
            pass
    else:
        if base_ok:
            return parent
        else:
            raise RuntimeError



import asyncio
import logging
from copy import copy
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Hashable,
    Literal,
    Optional,
    IO,
    Dict,
    Any,
    TYPE_CHECKING,
    Type,
    Generic,
    TypeVar,
    Union,
)

from slixmpp import JID, Iq, Message

from ..types import AvatarType
from ..util import SubclassableOnce

if TYPE_CHECKING:
    from .session import BaseSession


class LegacyContact(metaclass=SubclassableOnce):
    """
    This class represents a contact a gateway user can interact with.
    """

    RESOURCE: str = "slidge"
    """
    A resource is required for chat states (and maybe other stuff) to work properly.
    This is the name of the resource the contacts will use.
    """
    FEATURES = {
        "http://jabber.org/protocol/chatstates",
        "urn:xmpp:receipts",
        "vcard-temp",
        "jabber:x:oob",
        "urn:xmpp:message-correct:0",
    }
    """
    A list of features advertised through service discovery and client capabilities.
    """

    def __init__(
        self,
        session: "BaseSession",
        legacy_id: Hashable,
        jid_username: str,
    ):
        """
        :param session: The session this contact is part of
        :param legacy_id: The contact's legacy ID
        :param jid_username: User part of this contact's 'puppet' JID.
            NB: case-insensitive, and some special characters are not allowed
        """
        self.session = session
        self.user = session.user
        self.legacy_id = legacy_id
        self.jid_username = jid_username

        self._name: Optional[str] = None
        self._avatar: Optional[AvatarType] = None

        self.xmpp = session.xmpp
        asyncio.create_task(self.make_caps())
        asyncio.create_task(self.make_vcard())

    def __repr__(self):
        return f"<LegacyContact '{self.jid}' - '{self.name}' - {self.user}>"

    @property
    def jid(self) -> JID:
        """
        Full JID (including the 'puppet' resource) of the contact
        """
        j = JID(self.jid_username + "@" + self.xmpp.boundjid.bare)
        j.resource = self.RESOURCE
        return j

    @property
    def name(self):
        """
        Friendly name of the contact, as it should appear in the user's roster
        """
        return self._name

    @name.setter
    def name(self, n: Optional[str]):
        self._name = n

    @property
    def avatar(self):
        """
        An image that represents this contact
        """
        return self._avatar

    @avatar.setter
    def avatar(self, a: Optional[AvatarType]):
        self._avatar = a
        self.xmpp.loop.create_task(self.make_vcard())

    async def make_caps(self):
        """
        Configure slixmpp to correctly advertise this contact's capabilities.
        """
        jid = self.jid
        xmpp = self.xmpp

        xmpp["xep_0030"].add_identity(jid=jid, category="client", itype="bot")
        for f in self.FEATURES:
            await xmpp["xep_0030"].add_feature(feature=f, jid=jid)

        info = await xmpp["xep_0030"].get_info(jid, node=None, local=True)
        if isinstance(info, Iq):
            info = info["disco_info"]
        ver = xmpp["xep_0115"].generate_verstring(info, xmpp["xep_0115"].hash)
        await xmpp["xep_0030"].set_info(
            jid=jid,
            node="%s#%s" % (xmpp["xep_0115"].caps_node, ver),
            info=info,
        )

        await xmpp["xep_0115"].cache_caps(ver, info)
        await xmpp["xep_0115"].assign_verstring(jid, ver)

    async def make_vcard(self):
        """
        Configure slixmpp to correctly set this contact's vcard (in fact only its avatar ATM)
        """
        await self.xmpp.make_vcard(jid=self.jid, avatar=self.avatar)

    async def add_to_roster(self):
        """
        Add a contact to a user roster using :xep:`0356`
        """
        await self.xmpp["xep_0356"].set_roster(
            jid=self.user.jid,
            roster_items={
                self.jid.bare: {
                    "name": self.name,
                    "subscription": "both",
                    "groups": [self.xmpp.ROSTER_GROUP],
                }
            },
        )

    def online(self):
        """
        Send an "online" presence from this contact to the user.
        """
        self.xmpp.send_presence(
            pfrom=self.jid,
            pto=self.user.jid.bare,
        )

    def away(self):
        """
        Send an "away" presence from this contact to the user.

        Does not make much sense in the context of mobile, "always connected" network where
        :func:`slidge.LegacyContact.inactive` is probably more relevant.
        """
        self.xmpp.send_presence(pfrom=self.jid, pto=self.user.jid.bare, pshow="away")

    def busy(self):
        """
        Send a "busy" presence from this contact to the user.
        """
        self.xmpp.send_presence(pfrom=self.jid, pto=self.user.jid.bare, pshow="busy")

    def status(self, text: str):
        """
        Set a contact's status
        """
        self.xmpp.send_presence(pfrom=self.jid, pto=self.user.jid.bare, pstatus=text)

    def offline(self):
        """
        Send an "offline" presence from this contact to the user.
        """
        self.xmpp.send_presence(
            pfrom=self.jid, pto=self.user.jid.bare, ptype="unavailable"
        )

    def unsubscribe(self):
        """
        Send an "unsubscribed" presence from this contact to the user.
        """
        self.xmpp.send_presence(
            pfrom=self.jid, pto=self.user.jid.bare, ptype="unsubscribed"
        )

    def chat_state(self, state: str):
        msg = self.xmpp.make_message(mfrom=self.jid, mto=self.user.jid, mtype="chat")
        msg["chat_state"] = state
        msg.enable("no-store")
        msg.send()

    def active(self):
        """
        Send an "active" chat state (:xep:`0085`) from this contact to the user.
        """
        self.chat_state("active")

    def composing(self):
        """
        Send a "composing" (ie "typing notification") chat state (:xep:`0085`) from this contact to the user.
        """
        self.chat_state("composing")

    def paused(self):
        """
        Send a "paused" (ie "typing paused notification") chat state (:xep:`0085`) from this contact to the user.
        """
        self.chat_state("paused")

    def inactive(self):
        """
        Send an "inactive" (ie "typing paused notification") chat state (:xep:`0085`) from this contact to the user.
        """
        log.debug("%s go inactive", self)
        self.chat_state("inactive")

    def ack(self, legacy_msg_id: Hashable):
        """
        Send an "acknowledged" message marker (:xep:`0333`) from this contact to the user.

        :param legacy_msg_id: The message this marker refers to
        """
        self.send_marker(legacy_msg_id, "acknowledged")

    def received(self, legacy_msg_id: Hashable):
        """
        Send a "received" message marker (:xep:`0333`) and a "message delivery receipt"
        (:xep:`0184`)
        from this contact to the user

        :param legacy_msg_id: The message this marker refers to
        """
        self.send_marker(legacy_msg_id, "received")

    def displayed(self, legacy_msg_id: Hashable):
        """
        Send a "displayed" message marker (:xep:`0333`) from this contact to the user.

        :param legacy_msg_id: The message this marker refers to
        """
        self.send_marker(legacy_msg_id, "displayed")

    def send_marker(
        self,
        legacy_msg_id: Hashable,
        marker: Literal["acknowledged", "received", "displayed"],
    ):
        """
        Send a message marker (:xep:`0333`) from this contact to the user.

        NB: for the 'received' marker, this also sends a message receipt (:xep:`0184`)

        :param legacy_msg_id: ID of the message this marker refers to
        :param marker: The marker type

        """
        xmpp_id = self.session.sent.get(legacy_msg_id)
        if xmpp_id is None:
            log.debug("Cannot find the XMPP ID of this msg: %s", legacy_msg_id)
        else:
            if marker == "received":
                receipt = self.xmpp.Message()
                receipt["to"] = self.user.jid
                receipt["receipt"] = xmpp_id
                receipt["from"] = self.jid
                receipt.send()
            self.xmpp["xep_0333"].send_marker(
                mto=self.user.jid,
                id=xmpp_id,
                marker=marker,
                mfrom=self.jid,
            )

    def _make_message(self, **kwargs):
        return self.xmpp.make_message(mfrom=self.jid, mto=self.user.jid, **kwargs)

    def _send_message(self, msg: Message, legacy_msg_id: Optional[Any] = None):
        if legacy_msg_id is not None:
            msg.set_id(self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id))
        msg.send()

    def send_text(
        self,
        body: str = "",
        chat_state: Optional[str] = "active",
        legacy_msg_id: Optional[Hashable] = None,
    ):
        """
        Transmit a message from the contact to the user

        :param body: Context of the message
        :param chat_state: By default, will send an "active" chat state (:xep:`0085`) along with the
            message. Set this to ``None`` if this is not desired.
        :param legacy_msg_id: If you want to be able to transport read markers from the gateway
            user to the legacy network, specify this
        """
        msg = self._make_message(mbody=body, mtype="chat")
        if chat_state is not None:
            msg["chat_state"] = chat_state
        self._send_message(msg, legacy_msg_id)
        return msg

    async def send_file(
        self,
        filename: Union[Path, str],
        content_type: Optional[str] = None,
        input_file: Optional[IO[bytes]] = None,
    ):
        try:
            log.debug("HOST: %s", self.xmpp.server_host)
            url = await self.xmpp["xep_0363"].upload_file(
                filename=filename,
                content_type=content_type,
                input_file=input_file,
            )
        except Exception as e:
            log.exception(e)
        else:
            msg = self._make_message()
            msg["oob"]["url"] = url
            msg["body"] = url
            self._send_message(msg)

    def _carbon(self, msg: Message):
        carbon = Message()
        carbon["from"] = self.user.jid
        carbon["type"] = "chat"
        carbon["carbon_sent"] = msg

        from_ = copy(self.user.jid)
        from_.resource = "slidge"
        msg["from"] = from_
        for resource in self.xmpp.client_roster.presence(self.user.jid):
            to = copy(self.user.jid)
            to.resource = resource
            carbon["to"] = str(to)
            self.xmpp["xep_0356"].send_privileged_message(copy(carbon))

    def carbon(
        self,
        body: str,
        legacy_id: Optional[Any] = None,
        date: Optional[datetime] = None,
    ):
        """
        Sync a message sent from an official client by the gateway user to XMPP.

        Uses xep:`0356` to impersonate the XMPP user and send a carbon message.

        :param str body: Body of the message.
        :param legacy_id:
        :param str date: When was this message sent.
        """
        # we use Message() directly because we need xmlns="jabber:client"
        msg = Message()
        msg["to"] = self.jid.bare
        msg["type"] = "chat"
        msg["body"] = body
        if legacy_id:
            xmpp_id = self.session.legacy_msg_id_to_xmpp_msg_id(legacy_id)
            msg.set_id(xmpp_id)
            self.session.sent[legacy_id] = xmpp_id
        if date:
            if date.tzinfo is None:
                date = date.astimezone(timezone.utc)
            msg["delay"].set_stamp(date)

        self._carbon(msg)
        return msg.get_id()

    def carbon_read(self, legacy_msg_id: Any, date: Optional[datetime] = None):
        """
        Uses xep:`0356` to impersonate the XMPP user and send a carbon message.

        :param str legacy_msg_id:
        :param str date:
        """
        # we use Message() directly because we need xmlns="jabber:client"
        msg = Message()
        msg["to"] = self.jid.bare
        msg["type"] = "chat"
        msg["displayed"]["id"] = self.session.legacy_msg_id_to_xmpp_msg_id(
            legacy_msg_id
        )
        if date is not None:
            if date.tzinfo is None:
                date = date.astimezone(timezone.utc)
            msg["delay"].set_stamp(date)

        self._carbon(msg)

    def correct(self, legacy_msg_id: Any, new_text: str):
        msg = self._make_message()
        msg["replace"]["id"] = self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id)
        msg["body"] = new_text
        self._send_message(msg)


LegacyContactType = TypeVar("LegacyContactType", bound=LegacyContact)


class LegacyRoster(Generic[LegacyContactType], metaclass=SubclassableOnce):
    """
    Virtual roster of a gateway user, that allows to represent all
    of their contacts as singleton instances (if used properly and not too bugged).

    The point of having singletons is for slixmpp to correctly advertise
    capabilities and vcard of contacts.
    """

    def __init__(self, session: "BaseSession"):
        self._contact_cls: Type[
            LegacyContactType
        ] = LegacyContact.get_self_or_unique_subclass()
        self._contact_cls.xmpp = session.xmpp

        self.session = session
        self.contacts_by_bare_jid: Dict[str, LegacyContactType] = {}
        self.contacts_by_legacy_id: Dict[Any, LegacyContactType] = {}

    def __iter__(self):
        return iter(self.contacts_by_legacy_id.values())

    def by_jid(self, contact_jid: JID) -> LegacyContactType:
        """
        Retrieve a contact by their JID

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.jid_username_to_legacy_id` to infer their
        legacy user ID.

        :param contact_jid:
        :return:
        """
        bare = contact_jid.bare
        c = self.contacts_by_bare_jid.get(bare)
        if c is None:
            jid_username = str(contact_jid.username)
            log.debug("Contact %s not found", contact_jid)
            c = self._contact_cls(
                self.session,
                self.jid_username_to_legacy_id(jid_username),
                jid_username,
            )
            self.contacts_by_bare_jid[bare] = c
        return c

    def by_legacy_id(self, legacy_id: Any) -> LegacyContactType:
        """
        Retrieve a contact by their legacy_id

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.legacy_id_to_jid_username` to infer their
        legacy user ID.

        :param legacy_id:
        :return:
        """
        c = self.contacts_by_legacy_id.get(legacy_id)
        if c is None:
            log.debug("Contact %s not found in roster", legacy_id)
            c = self._contact_cls(
                self.session, legacy_id, self.legacy_id_to_jid_username(legacy_id)
            )
            self.contacts_by_legacy_id[legacy_id] = c
        return c

    def by_stanza(self, s) -> LegacyContactType:
        """
        Retrieve a contact by the destination of a stanza

        See :meth:`slidge.Roster.by_legacy_id` for more info.

        :param s:
        :return:
        """
        return self.by_jid(s.get_to())

    @staticmethod
    def legacy_id_to_jid_username(legacy_id: Any) -> str:
        """
        Convert a legacy ID to a valid 'user' part of a JID

        Should be overridden for cases where the str conversion of
        the legacy_id is not enough, e.g., if it contains forbidden character.

        :param legacy_id:
        """
        return str(legacy_id)

    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> Hashable:
        """
        Convert a JID user part to a legacy ID.

        Should be overridden in case legacy IDs are not strings, for instance

        :param jid_username:
        :return:
        """
        return jid_username


LegacyRosterType = TypeVar("LegacyRosterType", bound=LegacyRoster)

log = logging.getLogger(__name__)



"""
Gateway to the signal network, using signald. Only supports registering a new number currently.
Linking to an existing account will be implemented once file upload works.
"""
import datetime
import logging
from argparse import ArgumentParser
from typing import Dict, Optional, Hashable, List, Any

from slixmpp import Message, Presence, JID
from slixmpp.exceptions import XMPPError

from aiosignald import SignaldAPI
import aiosignald.generated as sigapi
import aiosignald.exc as sigexc

from slidge import *
from slidge.legacy.contact import LegacyContactType

from . import txt


class Gateway(BaseGateway):
    COMPONENT_NAME = "Signal (slidge)"
    COMPONENT_TYPE = "signal"
    COMPONENT_AVATAR = (
        "https://upload.wikimedia.org/wikipedia/commons/5/56/Logo_Signal..png"
    )
    REGISTRATION_INSTRUCTIONS = txt.REGISTRATION_INSTRUCTIONS
    REGISTRATION_FIELDS = txt.REGISTRATION_FIELDS

    ROSTER_GROUP = "Signal"

    def config(self, argv: List[str]):
        args = get_parser().parse_args(argv)

        global signald_socket
        signald_socket = args.socket
        self.loop.create_task(self.connect_signal())

    async def connect_signal(self, *_):
        """
        Establish connection to the signald socker
        """
        global signal
        log.debug("Connecting to signald...")
        _, signal = await self.loop.create_unix_connection(Signal, signald_socket)
        signal.xmpp = self

    async def on_gateway_message(self, msg: Message):
        log.debug("Gateway msg: %s", msg)
        user = user_store.get_by_stanza(msg)
        if user is None:
            raise XMPPError("Please register to the gateway first")
        try:
            f = self._input_futures.pop(user.bare_jid)
        except KeyError:
            cmd = msg["body"]
            if cmd == "add_device":
                await self.add_device(user)
            else:
                self.send_message(mto=msg.get_from(), mbody="Come again?", mtype="chat")
        else:
            f.set_result(msg["body"])

    async def add_device(self, user: GatewayUser):
        uri = await self.input(user, "URI?")
        session = Signal.sessions_by_phone[user.registration_form["phone"]]

        try:
            await signal.add_device(account=session.phone, uri=uri)
        except sigexc.SignaldException as e:
            self.send_message(mto=user.jid, mbody=f"Problem: {e}", mtype="chat")
        else:
            self.send_message(mto=user.jid, mbody=f"Linking OK", mtype="chat")

    async def validate(self, user_jid: JID, registration_form: Dict[str, str]):
        phone = registration_form.get("phone")
        for u in user_store.get_all():
            if u.registration_form.get("phone") == phone:
                raise XMPPError(
                    "not-allowed",
                    text="Someone is already using this phone number on this server.\n",
                )
        if registration_form.get("device") == "primary" and not registration_form.get(
            "name"
        ):
            raise ValueError(txt.NAME_REQUIRED)

    async def unregister(self, user: GatewayUser):
        try:
            await signal.delete_account(
                account=user.registration_form.get("phone"), server=False
            )
        except sigexc.NoSuchAccountError:
            # if user unregisters before completing the registration process,
            # NoSuchAccountError is raised by signald
            pass

        log.info("Removed user: %s", user)


# noinspection PyPep8Naming
class Signal(SignaldAPI):
    """
    Extends :class:`.SignaldAPI` with handlers for events we are interested in.
    """

    sessions_by_phone: Dict[str, "Session"] = {}

    @staticmethod
    async def handle_WebSocketConnectionState(
        state: sigapi.WebSocketConnectionStatev1, payload
    ):
        """
        Connection state for an account

        :param state: State of the connection
        :param payload: The raw payload sent by signald
        """
        phone = payload["account"]
        if state.state == "CONNECTED":
            session = Signal.sessions_by_phone[phone]
            await session.add_contacts_to_roster()

    @staticmethod
    async def handle_IncomingMessage(msg: sigapi.IncomingMessagev1, _payload):
        """
        Dispatch a signald message to the proper session.

        Can be a lot of other things than an actual message, still need to figure
        things out to cover all cases.

        :param msg: the data!
        :param _payload:
        """
        session = Signal.sessions_by_phone[msg.account]
        await session.on_signal_message(msg)


class Contact(LegacyContact):
    session: "Session"

    def __init__(
        self,
        session: "Session",
        phone: str,
        jid_username: str,
    ):
        super().__init__(session, phone, jid_username)
        log.debug("JID: %s", self.jid_username)
        self._uuid: Optional[str] = None

    @property
    def phone(self):
        return self.legacy_id

    @phone.setter
    def phone(self, p):
        if p is not None:
            self.session.contacts.contacts_by_legacy_id[p] = self
            self.legacy_id = p
            self.jid_username = p

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, u: str):
        if u is not None:
            log.debug("UUID: %s, %s", u, self)
            self.session.contacts.contacts_by_uuid[u] = self
        self._uuid = u

    @property
    def signal_address(self):
        return sigapi.JsonAddressv1(number=self.phone, uuid=self.uuid)


class Roster(LegacyRoster):
    session: "Session"
    contacts_by_legacy_id: Dict[str, Contact]

    def __init__(self, session):
        super().__init__(session)
        self.contacts_by_uuid: Dict[str, Contact] = {}

    def by_phone(self, phone: str):
        return self.by_legacy_id(phone)

    def by_uuid(self, uuid: str):
        try:
            return self.contacts_by_uuid[uuid]
        except KeyError:
            log.warning(f"Cannot find the contact corresponding to the UUID {uuid}")
            return Contact(self.session, "unknown_phone", "unknown_phone")

    def by_json_address(self, address: sigapi.JsonAddressv1):
        uuid = address.uuid
        phone = address.number

        if uuid is None and phone is None:
            raise TypeError(address)

        if uuid is None:
            return self.by_phone(phone)

        if phone is None:
            return self.by_uuid(uuid)

        contact_phone = self.contacts_by_legacy_id.get(phone)
        contact_uuid = self.contacts_by_uuid.get(uuid)

        if contact_phone is None and contact_uuid is None:
            c = self.by_phone(phone)
            c.uuid = uuid
            return c

        if contact_phone is None and contact_uuid is not None:
            contact_uuid.phone = phone
            return contact_uuid

        if contact_uuid is None and contact_phone is not None:
            contact_phone.uuid = uuid
            return contact_phone

        if contact_phone is not contact_uuid:
            raise RuntimeError(address, contact_phone, contact_uuid)

        return contact_phone


class Session(BaseSession):
    """
    Represents a signal account
    """

    contacts: Roster

    def __init__(self, user: GatewayUser):
        """

        :param user:
        """
        super().__init__(user)
        self.phone: str = self.user.registration_form["phone"]
        Signal.sessions_by_phone[self.phone] = self

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> int:
        try:
            return int(i)
        except ValueError:
            raise NotImplementedError

    async def paused(self, c: LegacyContactType):
        pass

    async def correct(self, text: str, legacy_msg_id: Any, c: LegacyContactType):
        pass

    async def search(self, form_values: Dict[str, str]):
        pass

    async def login(self, p: Presence = None):
        """
        Attempt to listen to incoming events for this account,
        or pursue the registration process if needed.
        """
        self.send_gateway_status("Connecting...", show="dnd")
        try:
            await signal.subscribe(account=self.phone)
        except sigexc.NoSuchAccountError:
            device = self.user.registration_form["device"]
            try:
                if device == "primary":
                    await self.register()
                elif device == "secondary":
                    await self.link()
                else:
                    # This should never happen
                    self.send_gateway_status("Disconnected", show="dnd")
                    raise TypeError("Unknown device type", device)
            except sigexc.SignaldException as e:
                self.xmpp.send_message(
                    mto=self.user.jid, mbody=f"Something went wrong: {e}"
                )
                raise
            await signal.subscribe(account=self.phone)

        self.send_gateway_status(f"Connected as {self.phone}")
        await self.add_contacts_to_roster()

    async def register(self):
        self.send_gateway_status("Registering…", show="dnd")
        try:
            await signal.register(self.phone)
        except sigexc.CaptchaRequiredError:
            self.send_gateway_status("Captcha required", show="dnd")
            captcha = await self.xmpp.input(self.user, txt.CAPTCHA_REQUIRED)
            await signal.register(self.phone, captcha=captcha)
        sms_code = await self.xmpp.input(
            self.user,
            f"Reply to this message with the code you have received by SMS at {self.phone}.",
        )
        await signal.verify(account=self.phone, code=sms_code)
        await signal.set_profile(
            account=self.phone, name=self.user.registration_form["name"]
        )
        self.send_gateway_message(txt.REGISTER_SUCCESS)

    async def send_linking_qrcode(self):
        self.send_gateway_status("QR scan needed", show="dnd")
        resp = await signal.generate_linking_uri()
        await self.send_qr(resp.uri)
        self.xmpp.send_message(
            mto=self.user.jid,
            mbody=f"Use this URI or QR code on another signal device to "
            f"finish linking your XMPP account\n{resp.uri}",
        )
        return resp

    async def link(self):
        resp = await self.send_linking_qrcode()
        try:
            await signal.finish_link(
                device_name=self.user.registration_form["device_name"],
                session_id=resp.session_id,
            )
        except sigexc.ScanTimeoutError:
            while True:
                r = await self.input(txt.LINK_TIMEOUT)
                if r in ("cancel", "link"):
                    break
                else:
                    self.send_gateway_message("Please reply either 'link' or 'cancel'")
            if r == "cancel":
                raise
            elif r == "link":
                await self.link()  # TODO: set a max number of attempts
        except sigexc.SignaldException as e:
            self.xmpp.send_message(
                mto=self.user.jid,
                mbody=f"Something went wrong during the linking process: {e}.",
            )
            raise
        else:
            self.send_gateway_message(txt.LINK_SUCCESS)

    async def logout(self, p: Optional[Presence]):
        pass

    async def add_contacts_to_roster(self):
        """
        Populate a user's roster
        """
        profiles = await signal.list_contacts(account=self.phone)
        for profile in profiles.profiles:
            full_profile = await signal.get_profile(
                account=self.phone, address=profile.address
            )
            contact = self.contacts.by_phone(profile.address.number)
            contact.uuid = profile.address.uuid
            contact.name = profile.name or profile.profile_name
            if full_profile.avatar is not None:
                with open(full_profile.avatar, "rb") as f:
                    contact.avatar = f.read()
            await contact.add_to_roster()
            contact.online()

    async def on_signal_message(self, msg: sigapi.IncomingMessagev1):
        """
        User has received 'something' from signal

        :param msg:
        """
        if msg.sync_message is not None:
            if msg.sync_message.sent is None:
                log.debug("Ignoring %s", msg)  # Probably a 'message read' marker
                return
            destination = msg.sync_message.sent.destination
            contact = self.contacts.by_json_address(destination)
            sent_msg = msg.sync_message.sent.message
            contact.carbon(
                body=sent_msg.body,
                date=datetime.datetime.fromtimestamp(sent_msg.timestamp / 1000),
            )

        contact = self.contacts.by_json_address(msg.source)

        if msg.data_message is not None:
            contact.send_text(
                body=msg.data_message.body,
                legacy_msg_id=msg.data_message.timestamp,
            )

        if msg.typing_message is not None:
            action = msg.typing_message.action
            if action == "STARTED":
                contact.active()
                contact.composing()
            elif action == "STOPPED":
                contact.paused()

        if msg.receipt_message is not None:
            type_ = msg.receipt_message.type
            if type_ == "DELIVERY":
                for t in msg.receipt_message.timestamps:
                    contact.received(t)
            elif type_ == "READ":
                for t in msg.receipt_message.timestamps:
                    contact.displayed(t)

    async def send_text(self, t: str, c: Contact) -> int:
        response = await signal.send(
            account=self.phone,
            recipientAddress=c.signal_address,
            messageBody=t,
        )
        result = response.results[0]
        log.debug("Result: %s", result)
        if (
            result.networkFailure
            or result.identityFailure
            or result.identityFailure
            or result.proof_required_failure
        ):
            raise XMPPError(str(result))
        return response.timestamp

    async def send_file(self, u: str, c: LegacyContact) -> Optional[Hashable]:
        pass

    async def active(self, c: LegacyContact):
        pass

    async def inactive(self, c: LegacyContact):
        pass

    async def composing(self, c: Contact):
        await signal.typing(
            account=self.phone,
            address=c.signal_address,
            typing=True,
        )

    async def displayed(self, legacy_msg_id: int, c: Contact):
        await signal.mark_read(
            account=self.phone,
            to=c.signal_address,
            timestamps=[legacy_msg_id],
        )


def get_parser():
    parser = ArgumentParser()
    parser.add_argument("--socket", default="/signald/signald.sock")
    return parser


log = logging.getLogger(__name__)

signal: Signal
signald_socket: Optional[str] = None



"""
A pseudo legacy network, to easily test things
"""

import asyncio
import logging
from pathlib import Path
from typing import Dict, Optional, Hashable, Any

from slixmpp import JID, Presence

from slidge import *
from slidge.legacy.contact import LegacyContactType

ASSETS_DIR = Path(__file__).parent.parent.parent / "assets"


class Gateway(BaseGateway):
    COMPONENT_NAME = "The great legacy network (slidge)"
    COMPONENT_AVATAR = ASSETS_DIR / "gateway.png"
    REGISTRATION_INSTRUCTIONS = (
        "Only username 'n' is accepted and only 'baba' and 'bibi' contacts exist.\n"
        "You can use any password you want."
    )
    REGISTRATION_FIELDS = list(BaseGateway.REGISTRATION_FIELDS) + [
        FormField(
            var="something_else",
            label="Some optional stuff not covered by jabber:iq:register",
            required=False,
            private=True,
        )
    ]

    async def validate(self, user_jid: JID, registration_form: Dict[str, str]):
        if registration_form["username"] != "n":
            raise ValueError("Y a que N!")


class Session(BaseSession):
    def __init__(self, user):
        super(Session, self).__init__(user)
        self.counter = 0

    async def paused(self, c: LegacyContactType):
        pass

    async def correct(self, text: str, legacy_msg_id: Any, c: LegacyContactType):
        pass

    async def login(self, p: Presence):
        log.debug("Logging in user: %s", p)
        self.send_gateway_status("Connecting...", show="dnd")
        await asyncio.sleep(1)
        self.send_gateway_status("Connected")
        for b, a in zip(BUDDIES, AVATARS):
            c = self.contacts.by_legacy_id(b.lower())
            c.name = b.title()
            c.avatar = a
            await c.add_to_roster()
            c.online()

    async def logout(self, p: Optional[Presence]):
        log.debug("User has disconnected")

    async def send_text(self, t: str, c: LegacyContact):
        i = self.counter
        self.counter = i + 1
        self.xmpp.loop.create_task(self.later(c))
        return i

    async def send_file(self, u: str, c: LegacyContact) -> Optional[Hashable]:
        i = self.counter
        self.counter = i + 1
        c.send_text(u)
        await c.send_file(ASSETS_DIR / "buddy1.png")
        return i

    async def later(self, c: LegacyContact):
        i = self.counter - 1
        await asyncio.sleep(1)
        c.received(i)
        await asyncio.sleep(1)
        c.active()
        await asyncio.sleep(1)
        c.ack(i)
        await asyncio.sleep(1)
        c.displayed(i)
        await asyncio.sleep(1)
        c.composing()
        await asyncio.sleep(1)
        c.paused()
        await asyncio.sleep(1)
        c.composing()
        await asyncio.sleep(1)
        c.send_text("OK", legacy_msg_id=i)
        await asyncio.sleep(1)
        c.inactive()

    async def active(self, c: LegacyContact):
        log.debug("User is active for contact %s", c)

    async def inactive(self, c: LegacyContact):
        log.debug("User is inactive for contact %s", c)

    async def composing(self, c: LegacyContact):
        log.debug("User is composing for contact %s", c)

    async def displayed(self, legacy_msg_id: int, c: LegacyContact):
        log.debug("Message #%s was read by the user", legacy_msg_id)

    async def search(self, form_values: Dict[str, str]):
        if form_values["first"] == "bubu":
            return SearchResult(
                fields=[FormField("first"), FormField("jid", type="jid-single")],
                items=[{"first": "bubu", "jid": f"bubu@{self.xmpp.boundjid.bare}"}],
            )


BUDDIES = ["baba", "bibi"]
AVATARS = []

with (ASSETS_DIR / "buddy1.png").open("rb") as fp:
    AVATARS.append(fp.read())

with (ASSETS_DIR / "buddy2.png").open("rb") as fp:
    AVATARS.append(fp.read())

log = logging.getLogger(__name__)



import asyncio
import datetime
import functools
import logging
import re
import tempfile
from argparse import Namespace
from pathlib import Path
from typing import Dict, List, Optional
from mimetypes import guess_type

import aiohttp
from slixmpp import JID, Presence
from slixmpp.exceptions import XMPPError

import aiotdlib
import aiotdlib.api as tgapi

from slidge import *
from .config import get_parser

REGISTRATION_INSTRUCTIONS = """You can visit https://my.telegram.org/apps to get an API ID and an API HASH

This is the only tested login method, but other methods (password, bot token, 2FA...)
should work too, in theory at least.
"""


class Gateway(BaseGateway):
    REGISTRATION_INSTRUCTIONS = REGISTRATION_INSTRUCTIONS
    REGISTRATION_FIELDS = [
        FormField(var="phone", label="Phone number", required=True),
        FormField(var="api_id", label="API ID", required=False),
        FormField(var="api_hash", label="API hash", required=False),
        FormField(var="", value="The fields below have not been tested", type="fixed"),
        FormField(var="bot_token", label="Bot token", required=False),
        FormField(var="first", label="First name", required=False),
        FormField(var="last", label="Last name", required=False),
    ]
    ROSTER_GROUP = "Telegram"
    COMPONENT_NAME = "Telegram (slidge)"
    COMPONENT_TYPE = "telegram"
    COMPONENT_AVATAR = "https://web.telegram.org/img/logo_share.png"

    SEARCH_FIELDS = [
        FormField(var="phone", label="Phone number", required=True),
    ]

    args: Namespace

    def config(self, argv: List[str]):
        Gateway.args = args = get_parser().parse_args(argv)
        if args.tdlib_path is None:
            args.tdlib_path = self.home_dir / "tdlib"

    async def validate(self, user_jid: JID, registration_form: Dict[str, str]):
        pass

    async def unregister(self, user):
        pass


class Contact(LegacyContact):
    legacy_id: int


class Roster(LegacyRoster):
    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> int:
        return int(jid_username)


class Session(BaseSession):
    tdlib_path: Optional[Path] = None
    tg: "TelegramClient"
    sent_read_marks: set[int]

    def post_init(self):
        registration_form = {
            k: v if v != "" else None for k, v in self.user.registration_form.items()
        }
        self.sent_read_marks = set()

        i = registration_form.get("api_id")
        if i is not None:
            i = int(i)  # makes testing easier to make api_id optional...

        self.tg = TelegramClient(
            self.xmpp,
            self,
            api_id=i,
            api_hash=registration_form.get("api_hash"),
            phone_number=registration_form["phone"],
            bot_token=registration_form.get("bot_token"),
            first_name=registration_form.get("first"),
            last_name=registration_form.get("last"),
            database_encryption_key=Gateway.args.tdlib_key,
            files_directory=Gateway.args.tdlib_path,
        )

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> int:
        try:
            return int(i)
        except ValueError:
            raise NotImplementedError("This is not a valid telegram msg ID")

    async def login(self, p: Presence):
        self.send_gateway_status("Connecting", show="dnd")
        async with self.tg as tg:
            self.send_gateway_status(f"Connected as {self.tg.get_my_id()}")
            await self.add_contacts_to_roster()
            await tg.idle()

    async def logout(self, p: Optional[Presence]):
        pass

    async def send_text(self, t: str, c: Contact) -> int:
        t = escape(t)
        try:
            result = await self.tg.send_text(chat_id=c.legacy_id, text=t)
        except tgapi.BadRequest as e:
            if e.code == 400:
                raise XMPPError(condition="item-not-found", text="No such contact")
            else:
                raise
        fut = self.xmpp.loop.create_future()
        ack_futures[result.id] = fut
        new_message_id = await fut
        log.debug("Result: %s / %s", result, new_message_id)
        return new_message_id

    async def send_file(self, u: str, c: Contact) -> int:
        type_, _ = guess_type(u)
        if type_ is not None:
            type_, subtype = type_.split("/")

        if type_ == "image":
            async with aiohttp.ClientSession() as session:
                async with session.get(u) as response:
                    response.raise_for_status()
                    with tempfile.NamedTemporaryFile() as file:
                        bytes_ = await response.read()
                        file.write(bytes_)
                        result = await self.tg.send_photo(
                            chat_id=c.legacy_id, photo=file.name
                        )
        else:
            result = await self.tg.send_text(chat_id=c.legacy_id, text=u)

        return result.id

    async def active(self, c: Contact):
        action = tgapi.OpenChat.construct(chat_id=c.legacy_id)
        res = await self.tg.request(action)
        log.debug("Open chat res: %s", res)

    async def inactive(self, c: Contact):
        action = tgapi.CloseChat.construct(chat_id=c.legacy_id)
        res = await self.tg.request(action)
        log.debug("Close chat res: %s", res)

    async def composing(self, c: Contact):
        action = tgapi.SendChatAction.construct(
            chat_id=c.legacy_id,
            action=tgapi.ChatActionTyping(),
            message_thread_id=0,  # TODO: check what telegram's threads really are
        )

        res = await self.tg.request(action)
        log.debug("Send composing res: %s", res)

    async def paused(self, c: Contact):
        pass

    async def displayed(self, tg_id: int, c: Contact):
        query = tgapi.ViewMessages.construct(
            chat_id=c.legacy_id,
            message_thread_id=0,
            message_ids=[tg_id],
            force_read=True,
        )
        res = await self.tg.request(query)
        log.debug("Send chat action res: %s", res)

    async def add_contacts_to_roster(self):
        chats = await self.tg.get_main_list_chats_all()
        for chat in chats:
            if not isinstance(chat.type_, tgapi.ChatTypePrivate):
                log.debug("Skipping %s as it is of type %s", chat.title, chat.type_)
            if isinstance(chat.photo, tgapi.ChatPhotoInfo):
                query = tgapi.DownloadFile.construct(
                    file_id=chat.photo.big.id, synchronous=True, priority=32
                )
                response: tgapi.File = await self.tg.request(query)
                with open(response.local.path, "rb") as f:
                    avatar = f.read()
            else:
                avatar = None
            contact = self.contacts.by_legacy_id(chat.id)
            contact.name = chat.title
            contact.avatar = avatar
            await contact.add_to_roster()
            contact.online()

    async def correct(self, text: str, legacy_msg_id: int, c: Contact):
        query = tgapi.EditMessageText.construct(
            chat_id=c.legacy_id,
            message_id=legacy_msg_id,
            input_message_content=tgapi.InputMessageText.construct(
                text=tgapi.FormattedText.construct(text=text)
            ),
        )
        await self.tg.request(query)

    async def search(self, form_values: Dict[str, str]):
        phone = form_values["phone"]
        response: tgapi.ImportedContacts = await self.tg.request(
            query=tgapi.ImportContacts(
                contacts=[
                    tgapi.Contact(
                        phone_number=phone,
                        user_id=0,
                        first_name=phone,
                        vcard="",
                        last_name="",
                    )
                ]
            )
        )
        user_id = response.user_ids[0]
        if user_id == 0:
            return

        await self.add_contacts_to_roster()
        contact: Contact = self.contacts.by_legacy_id(user_id)
        await contact.add_to_roster()

        return SearchResult(
            fields=[FormField("phone"), FormField("jid", type="jid-single")],
            items=[{"phone": form_values["phone"], "jid": contact.jid.bare}],
        )


class TelegramClient(aiotdlib.Client):
    def __init__(self, xmpp: BaseGateway, session: Session, **kw):
        super().__init__(parse_mode=aiotdlib.ClientParseMode.MARKDOWN, **kw)
        self.session = session

        async def input_(prompt):
            self.session.send_gateway_status(f"Action required: {prompt}")
            return await xmpp.input(session.user, prompt)

        self.input = input_
        self._auth_get_code = functools.partial(input_, "Enter code")
        self._auth_get_password = functools.partial(input_, "Enter 2FA password:")
        self._auth_get_first_name = functools.partial(input_, "Enter first name:")
        self._auth_get_last_name = functools.partial(input_, "Enter last name:")

        for h, t in [
            (on_telegram_message, tgapi.API.Types.UPDATE_NEW_MESSAGE),
            (on_message_success, tgapi.API.Types.UPDATE_MESSAGE_SEND_SUCCEEDED),
            (on_contact_status, tgapi.API.Types.UPDATE_USER_STATUS),
            (on_contact_chat_action, tgapi.API.Types.UPDATE_CHAT_ACTION),
            (on_contact_read, tgapi.API.Types.UPDATE_CHAT_READ_OUTBOX),
            (on_user_read_from_other_device, tgapi.API.Types.UPDATE_CHAT_READ_INBOX),
            (on_contact_edit_msg, tgapi.API.Types.UPDATE_MESSAGE_CONTENT),
            (on_user_update, tgapi.API.Types.UPDATE_USER),
        ]:
            self.add_event_handler(h, t)


async def on_telegram_message(tg: TelegramClient, update: tgapi.UpdateNewMessage):
    log.debug("Received message update")
    msg: tgapi.Message = update.message
    session = tg.session

    if msg.is_channel_post:
        log.debug("Ignoring channel post")
        return

    if msg.is_outgoing:
        # This means slidge is responsible for this message, so no carbon is needed;
        # but maybe this does not handle all possible cases gracefully?
        if msg.sending_state is not None or msg.id in session.sent:
            return
        contact = session.contacts.by_legacy_id(msg.chat_id)
        # noinspection PyUnresolvedReferences
        contact.carbon(
            msg.content.text.text, msg.id, datetime.datetime.fromtimestamp(msg.date)
        )
        return

    sender = msg.sender_id
    if not isinstance(sender, tgapi.MessageSenderUser):
        log.debug("Ignoring non-user sender")  # Does this happen?
        return

    contact = session.contacts.by_legacy_id(sender.user_id)

    content = msg.content
    if isinstance(content, tgapi.MessageText):
        # TODO: parse formatted text to markdown
        formatted_text = content.text
        contact.send_text(body=formatted_text.text, legacy_msg_id=msg.id)
        return

    if isinstance(content, tgapi.MessageAnimatedEmoji):
        emoji = content.animated_emoji.sticker.emoji
        contact.send_text(body=emoji, legacy_msg_id=msg.id)
        return

    if isinstance(content, tgapi.MessagePhoto):
        photo = content.photo
        best_file = max(photo.sizes, key=lambda x: x.width).photo
    elif isinstance(content, tgapi.MessageVideo):
        best_file = content.video.video
    else:
        raise NotImplemented

    query = tgapi.DownloadFile.construct(
        file_id=best_file.id, synchronous=True, priority=1
    )
    best_file_downloaded: tgapi.File = await tg.request(query)
    await contact.send_file(best_file_downloaded.local.path)
    if content.caption.text:
        contact.send_text(content.caption.text, legacy_msg_id=msg.id)


async def on_message_success(
    tg: TelegramClient, update: tgapi.UpdateMessageSendSucceeded
):
    tg.session.sent_read_marks.add(update.message.id)
    for _ in range(10):
        try:
            future = ack_futures.pop(update.message.id)
        except KeyError:
            await asyncio.sleep(0.5)
        else:
            future.set_result(update.message.id)
            return
    log.warning("Ignoring Send success for %s", update.message.id)


async def on_contact_status(tg: TelegramClient, update: tgapi.UpdateUserStatus):
    if update.user_id == await tg.get_my_id():
        return

    session = tg.session
    contact = session.contacts.by_legacy_id(update.user_id)
    status = update.status
    if isinstance(status, tgapi.UserStatusOnline):
        contact.active()
    elif isinstance(status, tgapi.UserStatusOffline):
        contact.paused()
        contact.inactive()
    else:
        log.debug("Ignoring status %s", update)


async def on_contact_read(tg: TelegramClient, update: tgapi.UpdateChatReadOutbox):
    tg.session.contacts.by_legacy_id(update.chat_id).displayed(
        update.last_read_outbox_message_id
    )


async def on_contact_chat_action(tg: TelegramClient, action: tgapi.UpdateChatAction):
    session = tg.session
    sender = action.sender_id
    if not isinstance(sender, tgapi.MessageSenderUser):
        log.debug("Ignoring action: %s", action)
        return

    chat_id = action.chat_id
    if chat_id != sender.user_id:
        log.debug("Ignoring action: %s", action)
        return
    contact = session.contacts.by_legacy_id(chat_id)
    contact.composing()


async def on_user_read_from_other_device(
    tg: TelegramClient, action: tgapi.UpdateChatReadInbox
):
    session = tg.session
    msg_id = action.last_read_inbox_message_id
    log.debug("Self read mark for %s and we sent %s", msg_id, session.sent_read_marks)
    try:
        session.sent_read_marks.remove(msg_id)
    except KeyError:
        # slidge didn't send this read mark, so it comes from the official tg client
        contact = session.contacts.by_legacy_id(action.chat_id)
        contact.carbon_read(msg_id)


async def on_contact_edit_msg(tg: TelegramClient, action: tgapi.UpdateMessageContent):
    new = action.new_content
    if not isinstance(new, tgapi.MessageText):
        raise NotImplementedError(new)
    session = tg.session
    contact = session.contacts.by_legacy_id(action.chat_id)
    contact.correct(action.message_id, new.text.text)


async def on_user_update(tg: TelegramClient, action: tgapi.UpdateUser):
    u = action.user
    if u.id == await tg.get_my_id():
        return
    await tg.request(
        query=tgapi.ImportContacts(
            contacts=[
                tgapi.Contact(
                    phone_number=u.phone_number,
                    user_id=u.id,
                    first_name=u.first_name,
                    last_name=u.last_name,
                    vcard="",
                )
            ]
        )
    )
    contact: Contact = tg.session.contacts.by_legacy_id(u.id)
    contact.name = u.first_name
    await contact.add_to_roster()


def escape(t: str):
    return re.sub(ESCAPE_PATTERN, r"\\\1", t)


RESERVED_CHARS = "_*[]()~`>#+-=|{}.!"
ESCAPE_PATTERN = re.compile(f"([{re.escape(RESERVED_CHARS)}])")


ack_futures: Dict[int, asyncio.Future] = {}
log = logging.getLogger(__name__)



from . import xep_0055
from . import xep_0077
from . import xep_0100
from . import xep_0115
from . import xep_0333
from . import xep_0356
from . import xep_0363

import slixmpp.plugins

from .db import user_store, GatewayUser
from .gateway import BaseGateway
from .util import FormField, SearchResult
from .legacy.session import BaseSession
from .legacy.contact import LegacyContact, LegacyRoster

slixmpp.plugins.__all__.extend(["xep_0055", "xep_0356"])

__all__ = [
    "BaseGateway",
    "BaseSession",
    "GatewayUser",
    "LegacyContact",
    "LegacyRoster",
    "FormField",
    "SearchResult",
    "user_store",
]



import functools
import logging
import random
import shelve
import time
from collections import defaultdict, deque
from pathlib import Path
from typing import Hashable, Dict, Union, Deque, Optional

import aiohttp
import maufbapi.types.graphql
from maufbapi import AndroidAPI, AndroidMQTT, AndroidState
from maufbapi.types import mqtt as mqtt_t

from slixmpp import Presence, JID
from slixmpp.exceptions import XMPPError

from slidge import *


class Gateway(BaseGateway):
    REGISTRATION_INSTRUCTIONS = "Enter facebook credentials"
    REGISTRATION_FIELDS = [
        FormField(var="email", label="Email", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]

    ROSTER_GROUP = "Facebook"

    COMPONENT_NAME = "Facebook (slidge)"
    COMPONENT_TYPE = "facebook"
    COMPONENT_AVATAR = "https://upload.wikimedia.org/wikipedia/commons/thumb/6/6c/Facebook_Messenger_logo_2018.svg/480px-Facebook_Messenger_logo_2018.svg.png"

    async def validate(self, user_jid: JID, registration_form: Dict[str, str]):
        pass


class Contact(LegacyContact):
    legacy_id: int


class Roster(LegacyRoster[Contact]):
    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> int:
        return int(jid_username)


class Session(BaseSession[Contact, Roster]):
    fb_state: AndroidState

    shelf_path: Path
    mqtt: AndroidMQTT
    api: AndroidAPI

    me: maufbapi.types.graphql.OwnInfo
    sent_messages: "Messages"
    received_messages: "Messages"

    contacts: Roster

    def post_init(self):
        self.shelf_path = self.xmpp.home_dir / self.user.bare_jid
        self.sent_messages = Messages()
        self.received_messages = Messages()

    async def login(self, p: Presence):
        shelf: shelve.Shelf[AndroidState]
        with shelve.open(str(self.shelf_path)) as shelf:
            try:
                # noinspection PyTypeChecker
                self.fb_state = s = shelf["state"]
            except KeyError:
                s = AndroidState()
                self.api = api = AndroidAPI(state=s)
                s.generate(random.randbytes(30))  # type: ignore
                await api.mobile_config_sessionless()
                try:
                    login = await api.login(
                        email=self.user.registration_form["email"],
                        password=self.user.registration_form["password"],
                    )
                except maufbapi.http.errors.IncorrectPassword:
                    self.send_gateway_message("Incorrect password")
                    raise
                except maufbapi.http.errors.TwoFactorRequired:
                    code = await self.input(
                        "Reply to this message with your 2 factor authentication code"
                    )
                    login = await api.login_2fa(
                        email=self.user.registration_form["email"], code=code
                    )
                log.debug(login)
                self.fb_state = shelf["state"] = api.state
            else:
                # noinspection PyTypeCheckers
                self.api = api = AndroidAPI(state=s)
        self.send_gateway_message("Login successful")
        self.mqtt = AndroidMQTT(api.state)
        self.me = await self.api.get_self()
        await self.add_friends()
        self.mqtt.seq_id_update_callback = lambda i: setattr(self.mqtt, "seq_id", i)
        self.mqtt.add_event_handler(mqtt_t.Message, self.on_fb_message)
        self.mqtt.add_event_handler(mqtt_t.ExtendedMessage, self.on_fb_message)
        self.mqtt.add_event_handler(mqtt_t.ReadReceipt, self.on_fb_message_read)
        self.mqtt.add_event_handler(mqtt_t.TypingNotification, self.on_fb_typing)
        self.mqtt.add_event_handler(mqtt_t.OwnReadReceipt, self.on_fb_user_read)

        self.mqtt.add_event_handler(mqtt_t.NameChange, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.AvatarChange, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.UnsendMessage, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.Reaction, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.Presence, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.AddMember, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.RemoveMember, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.ThreadChange, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.MessageSyncError, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.ForcedFetch, self.on_fb_event)
        # self.mqtt.add_event_handler(Connect, self.on_connect)
        # self.mqtt.add_event_handler(Disconnect, self.on_disconnect)
        await self.mqtt.listen(self.mqtt.seq_id)

    async def add_friends(self):
        thread_list = await self.api.fetch_thread_list(msg_count=0)
        self.mqtt.seq_id = int(thread_list.sync_sequence_id)
        log.debug("SEQ ID: %s", self.mqtt.seq_id)
        for t in thread_list.nodes:
            if t.is_group_thread:
                log.debug("Skipping group: %s", t)
                continue

            if len(t.all_participants.nodes) != 2:
                raise RuntimeError(t.all_participants)

            for participant in t.all_participants.nodes:
                if participant.id != self.me.id:
                    break
            else:
                raise RuntimeError(t.all_participants)

            c = self.contacts.by_legacy_id(int(participant.id))
            c.name = participant.messaging_actor.name
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    participant.messaging_actor.profile_pic_large.uri
                ) as response:
                    response.raise_for_status()
                    c.avatar = await response.read()
            await c.add_to_roster()
            log.debug("Contact: %s", c)
            c.online()

    async def logout(self, p: Optional[Presence]):
        pass

    async def send_text(self, t: str, c: Contact) -> int:
        resp: mqtt_t.SendMessageResponse = await self.mqtt.send_message(
            target=c.legacy_id, message=t, is_group=False
        )
        timestamp = get_now_ms()
        log.debug("Send message response: %s", resp)
        if not resp.success:
            raise XMPPError(resp.error_message)
        self.sent_messages.add(c.legacy_id, timestamp)
        return timestamp

    async def send_file(self, u: str, c: Contact) -> int:
        pass

    async def active(self, c: Contact):
        pass

    async def inactive(self, c: Contact):
        pass

    async def composing(self, c: Contact):
        await self.mqtt.set_typing(target=c.legacy_id)

    async def paused(self, c: Contact):
        await self.mqtt.set_typing(target=c.legacy_id, typing=False)

    async def displayed(self, legacy_msg_id: Hashable, c: Contact):
        await self.mqtt.mark_read(
            target=c.legacy_id, read_to=get_now_ms(), is_group=False
        )

    async def on_fb_message(self, evt: Union[mqtt_t.Message, mqtt_t.ExtendedMessage]):
        meta = evt.metadata
        log.debug("Facebook message: %s", evt)
        if str(meta.sender) == self.me.id:
            if f"app_id:{self.mqtt.state.application.client_id}" in meta.tags:
                log.debug("Ignoring self message")
            else:
                c = self.contacts.by_legacy_id(meta.thread.other_user_id)
                t = get_now_ms()
                c.carbon(body=evt.text, legacy_id=t)
                self.sent_messages.add(c.legacy_id, t)
            return

        t = get_now_ms()
        c = self.contacts.by_legacy_id(meta.sender)
        self.received_messages.add(c.legacy_id, t)
        c.send_text(evt.text, legacy_msg_id=t)

    async def on_fb_message_read(self, receipt: mqtt_t.ReadReceipt):
        log.debug("Facebook read: %s", receipt)
        try:
            real_timestamp = self.sent_messages.find_closest(
                receipt.user_id, receipt.read_to
            )
        except KeyError:
            log.debug("Could not find message with corresponding timestamp, ignoring")
        else:
            self.contacts.by_legacy_id(receipt.user_id).displayed(real_timestamp)

    async def on_fb_typing(self, notification: mqtt_t.TypingNotification):
        c = self.contacts.by_legacy_id(notification.user_id)
        if notification.typing_status:
            c.composing()
        else:
            c.paused()

    async def on_fb_user_read(self, receipt: mqtt_t.OwnReadReceipt):
        when = receipt.read_to
        for thread in receipt.threads:
            c = self.contacts.by_legacy_id(thread.other_user_id)
            try:
                timestamp = self.received_messages.find_closest(c.legacy_id, when)
            except KeyError:
                log.debug("Cannot find message to carbon read, ignoring")
                continue
            c.carbon_read(timestamp)

    @staticmethod
    async def on_fb_event(evt):
        log.debug("Facebook event: %s", evt)


class Messages:
    MAX_LENGTH = 500

    def __init__(self):
        self._messages: Dict[int, Deque[int]] = defaultdict(
            functools.partial(deque, maxlen=self.MAX_LENGTH)
        )

    def add(self, contact_id: int, timestamp_ms: int):
        self._messages[contact_id].append(timestamp_ms)

    def find_closest(self, contact_id: int, approx_timestamp_ms: int) -> int:
        messages = self._messages[contact_id]
        t: Optional[int] = None
        while True:
            if len(messages) == 0:
                if t is None:
                    raise KeyError(contact_id, approx_timestamp_ms)
                else:
                    return t
            peek = messages[0]
            if peek < approx_timestamp_ms:
                t = messages.popleft()
            else:
                if t is None:
                    raise KeyError(contact_id, approx_timestamp_ms)
                return t


def get_now_ms():
    return time.time_ns() // 1_000_000


log = logging.getLogger(__name__)



from argparse import ArgumentParser


def get_parser():
    parser = ArgumentParser()
    parser.add_argument("--tdlib-path", help="Defaults to ${SLIDGE_HOME_DIR}/tdlib")
    parser.add_argument(
        "--tdlib-key",
        default="NOT_SECURE",
        help="Key used to encrypt tdlib persistent DB",
    )
    return parser



# slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
import logging
from typing import Optional

from slixmpp import Message, JID
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0333 import stanza, Markable, Received, Displayed, Acknowledged

log = logging.getLogger(__name__)


class XEP_0333(BasePlugin):

    name = "xep_0333"
    description = "XEP-0333: Chat Markers (slidge)"
    stanza = stanza
    dependencies = {"xep_0030"}

    def plugin_init(self):
        register_stanza_plugin(Message, Markable)
        register_stanza_plugin(Message, Received)
        register_stanza_plugin(Message, Displayed)
        register_stanza_plugin(Message, Acknowledged)

        self.xmpp.register_handler(
            Callback(
                "Received Chat Marker",
                StanzaPath("message/received"),
                self._handle_received,
            )
        )
        self.xmpp.register_handler(
            Callback(
                "Displayed Chat Marker",
                StanzaPath("message/displayed"),
                self._handle_displayed,
            )
        )
        self.xmpp.register_handler(
            Callback(
                "Acknowledged Chat Marker",
                StanzaPath("message/acknowledged"),
                self._handle_acknowledged,
            )
        )

    def session_bind(self, jid):
        self.xmpp.plugin["xep_0030"].add_feature(stanza.NS)

    def plugin_end(self):
        self.xmpp.plugin["xep_0030"].del_feature(feature=stanza.NS)

    def _handle_received(self, message):
        self.xmpp.event("marker_received", message)
        self.xmpp.event("marker", message)

    def _handle_displayed(self, message):
        self.xmpp.event("marker_displayed", message)
        self.xmpp.event("marker", message)

    def _handle_acknowledged(self, message):
        self.xmpp.event("marker_acknowledged", message)
        self.xmpp.event("marker", message)

    def send_marker(
        self,
        mto: JID,
        id: str,
        marker: str,
        thread: Optional[str] = None,
        *,
        mfrom: Optional[JID] = None
    ):
        """
        Send a chat marker.

        :param JID mto: recipient of the marker
        :param str id: Identifier of the marked message
        :param str marker: Marker to send (one of
            displayed, received, or acknowledged)
        :param str thread: Message thread
        :param str mfrom: Use a specific JID to send the message
        """
        if marker not in ("displayed", "received", "acknowledged"):
            raise ValueError("Invalid marker: %s" % marker)
        msg = self.xmpp.make_message(mto=mto, mfrom=mfrom)
        if thread:
            msg["thread"] = thread
        msg[marker]["id"] = id
        msg.send()



"""
This module extends slixmpp.ComponentXMPP to make writing new LegacyClients easier
"""
import hashlib
import logging
import re
import tempfile
from asyncio import Future
from pathlib import Path
from typing import Dict, Iterable, Optional, List, Any

import aiohttp
import qrcode
from slixmpp import ComponentXMPP, Message, Iq, JID, Presence
from slixmpp.exceptions import XMPPError
from slixmpp.types import MessageTypes

from .db import user_store, RosterBackend, GatewayUser
from .legacy.session import BaseSession
from .util import FormField, SearchResult, ABCSubclassableOnceAtMost
from .types import AvatarType


class BaseGateway(ComponentXMPP, metaclass=ABCSubclassableOnceAtMost):
    """
    Class responsible for interacting with the gateway user ((un)registration) and dispatching
    messages from the user (or any slixmpp event) to the appropriate handlers.
    """

    REGISTRATION_FIELDS: Iterable[FormField] = [
        FormField(var="username", label="User name", required=True),
        FormField(var="password", label="Password", required=True),
    ]
    """
    Iterable of fields presented to the gateway user when registering using :xep:`0077`
    `extended <https://xmpp.org/extensions/xep-0077.html#extensibility>`_ by :xep:`0004`
    """
    REGISTRATION_INSTRUCTIONS: str = "Enter your credentials"

    COMPONENT_NAME: str = NotImplemented
    """Name of the component, as seen in service discovery"""
    COMPONENT_TYPE: Optional[str] = ""
    """Type of the gateway, should ideally follow https://xmpp.org/registrar/disco-categories.html"""
    COMPONENT_AVATAR: Optional[AvatarType] = None

    ROSTER_GROUP: str = "slidge"

    SEARCH_FIELDS: Iterable[FormField] = [
        FormField(var="first", label="First name", required=True),
        FormField(var="last", label="Last name", required=True),
    ]
    SEARCH_TITLE: str = "Search for legacy contacts"
    SEARCH_INSTRUCTIONS: str = ""

    def __init__(self, args):
        """

        :param args: CLI arguments parsed by :func:`.slidge.__main__.get_parser()`
        """
        super().__init__(
            args.jid,
            args.secret,
            args.server,
            args.port,
            plugin_whitelist=SLIXMPP_PLUGINS,
            plugin_config={
                "xep_0077": {
                    "form_fields": None,
                    "form_instructions": self.REGISTRATION_INSTRUCTIONS,
                },
                "xep_0100": {
                    "component_name": self.COMPONENT_NAME,
                    "user_store": user_store,
                    "type": self.COMPONENT_TYPE,
                },
                "xep_0184": {
                    "auto_ack": False,
                    "auto_request": True,
                },
                "xep_0363": {
                    "upload_service": args.upload_service,
                },
            },
        )
        self.home_dir = Path(args.home_dir)
        self._jid_validator = re.compile(args.user_jid_validator)
        self._config = args

        self._session_cls = BaseSession.get_unique_subclass()
        self._session_cls.xmpp = self

        self.register_plugins()
        self._register_slixmpp_api()
        self._register_handlers()
        self._input_futures: Dict[str, Future] = {}

    def _register_slixmpp_api(self):
        self["xep_0077"].api.register(
            user_store.get,
            "user_get",
        )
        self["xep_0077"].api.register(
            user_store.remove,
            "user_remove",
        )
        self["xep_0077"].api.register(
            self.make_registration_form,
            "make_registration_form",
        )
        self["xep_0077"].api.register(self._user_validate, "user_validate")
        self["xep_0077"].api.register(self._user_modify, "user_modify")

        self["xep_0055"].api.register(self._search_get_form, "search_get_form")
        self["xep_0055"].api.register(self._search_query, "search_query")

        self.roster.set_backend(RosterBackend)

    def _register_handlers(self):
        self.add_event_handler("session_start", self.on_session_start)
        self.add_event_handler("gateway_message", self.on_gateway_message)
        self.add_event_handler("user_register", self._on_user_register)
        self.add_event_handler("user_unregister", self._on_user_unregister)
        self.add_event_handler("legacy_login", self._legacy_login)
        get_session = self._session_cls.from_stanza

        # fmt: off
        async def logout(p): await get_session(p).logout(p)
        async def msg(m): await get_session(m).send_from_msg(m)
        async def disp(m): await get_session(m).displayed_from_msg(m)
        async def active(m): await get_session(m).active_from_msg(m)
        async def inactive(m): await get_session(m).inactive_from_msg(m)
        async def composing(m): await get_session(m).composing_from_msg(m)
        async def paused(m): await get_session(m).paused_from_msg(m)
        async def correct(m): await get_session(m).correct_from_msg(m)
        # fmt: on

        self.add_event_handler("legacy_logout", logout)
        self.add_event_handler("legacy_message", msg)
        self.add_event_handler("marker_displayed", disp)
        self.add_event_handler("chatstate_active", active)
        self.add_event_handler("chatstate_inactive", inactive)
        self.add_event_handler("chatstate_composing", composing)
        self.add_event_handler("chatstate_paused", paused)
        self.add_event_handler("message_correction", correct)

    def _add_commands(self):
        self["xep_0050"].add_command(
            node="info", name="Server Information", handler=self._handle_info
        )

    def _handle_info(self, iq: Iq, session: Dict[str, Any]):
        if iq.get_from().bare not in self._config.admins:
            raise XMPPError("not-authorized")
        form = self["xep_0004"].make_form("result", "Component info")
        form.add_field(
            ftype="jid-multi",
            label="Users",
            value=[u.bare_jid for u in user_store.get_all()],
        )

        session["payload"] = form
        session["has_next"] = False

        return session

    async def _validate_form(self, ifrom, form_dict):
        for field in self.REGISTRATION_FIELDS:
            if field.required and not form_dict.get(field.var):
                raise ValueError(f"Missing field: '{field.label}'")

        await self.validate(ifrom, form_dict)

    async def _user_validate(
        self, _gateway_jid, _node, ifrom: JID, form_dict: Dict[str, str]
    ):
        log.debug("User validate: %s", ifrom.bare)
        if not self._jid_validator.match(ifrom.bare):
            raise XMPPError(condition="not-allowed")
        await self._validate_form(ifrom, form_dict)
        log.info("New user: %s", ifrom.bare)
        user_store.add(ifrom, form_dict)

    async def _legacy_login(self, p: Presence):
        """
        Logs a :class:`slidge.BaseSession` instance to the legacy network

        :param p: Presence from a :class:`slidge.GatewayUser` directed at the gateway's own JID
        """
        session = self._session_cls.from_stanza(p)
        if not session.logged:
            session.logged = True
            await session.login(p)
            log.info("User logged in: %s", p.get_from().bare)

    async def _user_modify(
        self, _gateway_jid, _node, ifrom: JID, form_dict: Dict[str, str]
    ):
        user = user_store.get_by_jid(ifrom)
        log.debug("Modify user: %s", user)
        await self._validate_form(ifrom, form_dict)
        user_store.add(ifrom, form_dict)

    async def _on_user_unregister(self, iq: Iq):
        await self._session_cls.kill_by_jid(iq.get_from())

    async def _on_user_register(self, iq: Iq):
        for jid in self._config.admins:
            self.send_message(
                mto=jid, mbody=f"{iq.get_from()} has registered", mtype="headline"
            )

    async def _search_get_form(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError(text="Search is only allowed for registered users")

        reply = iq.reply()
        form = reply["search"]["form"]
        form["title"] = self.SEARCH_TITLE
        form["instructions"] = self.SEARCH_INSTRUCTIONS
        for field in self.SEARCH_FIELDS:
            form.add_field(**field.dict())
        return reply

    async def _search_query(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError(text="Search is only allowed for registered users")

        result: SearchResult = await self._session_cls.from_stanza(iq).search(
            iq["search"]["form"].get_values()
        )

        if not result:
            raise XMPPError("item-not-found", text="Nothing was found")

        reply = iq.reply()
        form = reply["search"]["form"]
        for field in result.fields:
            form.add_reported(field.var, label=field.label, type=field.type)
        for item in result.items:
            form.add_item(item)
        return reply

    def config(self, argv: List[str]):
        """
        Override this to access CLI args to configure the slidge plugin

        :param argv: CLI args that were not parsed by Slidge
        """
        pass

    async def validate(self, user_jid: JID, registration_form: Dict[str, str]):
        """
        Validate a registration form from a user.

        Since :xep:`0077` is pretty limited in terms of validation, it is OK to validate
        anything that looks good here and continue the legacy auth process via direct messages
        to the user (using :func:`.BaseGateway.input` for instance)

        :param user_jid:
        :param registration_form:
        """
        pass

    async def unregister(self, user: GatewayUser):
        """
        Optionally override this if you need to clean additional
        stuff after a user has been removed from the user_store.

        :param user:
        :return:
        """
        pass

    async def on_session_start(self, event):
        log.debug("Gateway session start: %s", event)
        # prevents XMPP clients from considering the gateway as an HTTP upload
        await self["xep_0030"].del_feature(
            feature="urn:xmpp:http:upload:0", jid=self.boundjid.bare
        )
        for user in user_store.get_all():
            # We need to see which registered users are online, this will trigger legacy_login in return
            self["xep_0100"].send_presence(ptype="probe", pto=user.jid)
        self._add_commands()
        await self.make_vcard(self.boundjid.bare, self.COMPONENT_AVATAR)
        log.info("Slidge has successfully started")

    async def make_vcard(self, jid: JID, avatar: bytes):
        """
        Configure slixmpp to correctly set this contact's vcard (in fact only its avatar ATM)
        """
        vcard = self["xep_0054"].make_vcard()
        if avatar is not None:
            if isinstance(avatar, bytes):
                avatar_bytes = avatar
            elif isinstance(avatar, Path):
                with avatar.open("rb") as f:
                    avatar_bytes = f.read()
            elif isinstance(avatar, str):
                async with aiohttp.ClientSession() as session:
                    async with session.get(avatar) as response:
                        avatar_bytes = await response.read()
            else:
                raise TypeError("Avatar must be bytes, a Path or a str (URL)", avatar)
            vcard["PHOTO"]["BINVAL"] = avatar_bytes
            await self["xep_0153"].api["set_hash"](
                jid=jid, args=hashlib.sha1(avatar_bytes).hexdigest()
            )
        await self["xep_0054"].api["set_vcard"](
            jid=jid,
            args=vcard,
        )

    def shutdown(self):
        log.debug("Shutting down")
        for user in user_store.get_all():
            session = self._session_cls.from_jid(user.jid)
            for c in session.contacts:
                c.offline()
            self["xep_0100"].send_presence(ptype="unavailable", pto=user.jid)

    async def make_registration_form(self, _jid, _node, _ifrom, iq: Iq):
        if not self._jid_validator.match(iq.get_from().bare):
            raise XMPPError(condition="not-allowed")

        reg = iq["register"]
        user = user_store.get_by_stanza(iq)
        log.debug("User found: %s", user)

        form = reg["form"]
        form.add_field(
            "FORM_TYPE",
            ftype="hidden",
            value="jabber:iq:register",
        )
        form["title"] = f"Registration to '{self.COMPONENT_NAME}'"
        form["instructions"] = self.REGISTRATION_INSTRUCTIONS

        if user is not None:
            reg["registered"] = False
            form.add_field(
                "remove",
                label="Remove my registration",
                required=True,
                ftype="boolean",
                value=False,
            )

        for field in self.REGISTRATION_FIELDS:
            if field.var in reg.interfaces:
                val = None if user is None else user.get(field.var)
                if val is None:
                    reg.add_field(field.var)
                else:
                    reg[field.var] = val

        reg["instructions"] = self.REGISTRATION_INSTRUCTIONS

        for field in self.REGISTRATION_FIELDS:
            form.add_field(
                field.var,
                label=field.label,
                required=field.required,
                ftype=field.type,
                options=field.options,
                value=field.value if user is None else user.get(field.var, field.value),
            )

        reply = iq.reply()
        reply.set_payload(reg)
        return reply

    async def on_gateway_message(self, msg: Message):
        """
        Called when an XMPP user (not necessarily registered as a gateway user) sends a direct message to
        the gateway.

        If you override this and still want :func:`.input` to work, make sure to include the try/except part.

        :param msg: Message sent by the XMPP user
        """
        log.debug("Gateway msg: %s", msg)
        user = user_store.get_by_stanza(msg)
        if user is None:
            return
        try:
            f = self._input_futures.pop(user.bare_jid)
        except KeyError:
            r = msg.reply(body="I got that, but I'm not doing anything with it")
            r["type"] = "chat"
            self.send(r)
        else:
            f.set_result(msg["body"])

    async def input(
        self, user: GatewayUser, text=None, mtype: MessageTypes = "chat", **msg_kwargs
    ):
        """
        Request arbitrary user input using simple message stanzas, and await the result.

        :param user: The (registered) user we want input from
        :param text: A prompt to display for the user
        :param mtype:
        :return:
        """
        if text is not None:
            self.send_message(mto=user.jid, mbody=text, mtype=mtype, **msg_kwargs)
        f = self.loop.create_future()
        self._input_futures[user.bare_jid] = f
        await f
        return f.result()

    async def send_file(self, filename: str, **msg_kwargs):
        url = await self["xep_0363"].upload_file(filename=filename)
        msg = self.make_message(**msg_kwargs)
        msg["oob"]["url"] = url
        msg["body"] = url
        msg.send()

    async def send_qr(self, text: str, **msg_kwargs):
        qr = qrcode.make(text)
        with tempfile.NamedTemporaryFile(suffix=".png") as f:
            qr.save(f.name)
            await self.send_file(f.name, **msg_kwargs)


SLIXMPP_PLUGINS = [
    "xep_0050",  # Adhoc commands
    "xep_0054",  # vCard-temp
    "xep_0055",  # Jabber search
    "xep_0066",  # Out of Band Data
    "xep_0077",  # In-band registration
    "xep_0085",  # Chat state notifications
    "xep_0100",  # Gateway interaction
    "xep_0115",  # Entity capabilities
    "xep_0153",  # vCard-Based Avatars
    "xep_0184",  # Message Delivery Receipts
    "xep_0280",  # Carbons
    "xep_0308",  # Last message correction
    "xep_0333",  # Chat markers
    "xep_0334",  # Message Processing Hints
    "xep_0356",  # Privileged Entity  (different registration because not listed in slixmpp.plugins.__all__
    "xep_0363",  # HTTP file upload
]
log = logging.getLogger(__name__)



"""
Hackernews slidge plugin

Will poll replies to items you've posted.
For every reply, the chat window '<REPLY_ITEM_ID>@<BRIDGE_JID>' should open.
It will contain your original post (as a carbon) and its reply as a normal chat message.
You can re-reply by replying in your XMPP client directly.
"""

import asyncio
import logging
import re
from datetime import datetime
from typing import Dict, Any, Optional

import aiohttp
from slixmpp import JID, Presence
from slixmpp.exceptions import XMPPError

from slidge import *


class Gateway(BaseGateway):
    # FIXME: implement proper login process, but we might have to do something to handle captcha
    REGISTRATION_INSTRUCTIONS = (
        "Enter the hackernews cookie from your browser's dev console "
        "(something like your-user-name ampersand XXXXXXXXXXXXXXXXXXXXXXXXX)"
    )
    REGISTRATION_FIELDS = [
        FormField(var="cookie", label="'user' cookie", required=True),
    ]

    ROSTER_GROUP = "HN"  # Not used, we don't add anything to the roster

    COMPONENT_NAME = "Hackernews (slidge)"
    COMPONENT_TYPE = "hackernews"

    COMPONENT_AVATAR = "https://news.ycombinator.com/favicon.ico"

    async def validate(self, user_jid: JID, registration_form: dict[str, str]):
        async with aiohttp.ClientSession(
            cookies={"user": registration_form["cookie"]}
        ) as session:
            async with session.get(LOGIN_URL, allow_redirects=False) as r:
                log.debug("Login response: %s - %s", r.status, await r.text())
                if r.status != 302:
                    raise ValueError("Cookie does not seem valid")


class Session(BaseSession[LegacyContact, LegacyRoster]):
    http_session: aiohttp.ClientSession
    highest_handled_submission_id: int
    hn_username: str

    def post_init(self):
        self.http_session = aiohttp.ClientSession(
            cookies={"user": self.user.registration_form["cookie"]}
        )
        self.highest_handled_submission_id = 0
        self.hn_username = self.user.registration_form["cookie"].split("&")[0]

    async def login(self, p: Presence):
        kid_ids: list[int] = []
        for submission_id in await self.get_user_submissions():
            user_submission = await self.get_item(submission_id)
            for kid_id in user_submission.get("kids", []):
                kid_ids.append(kid_id)

        if kid_ids:
            self.highest_handled_submission_id = max(kid_ids)

        while True:
            kid_ids.clear()
            for submission_id in await self.get_user_submissions():
                user_submission = await self.get_item(submission_id)
                for kid_id in user_submission.get("kids", []):
                    if kid_id <= self.highest_handled_submission_id:
                        continue
                    await self.send_own_and_reply(user_submission, kid_id)
                    kid_ids.append(kid_id)
            if kid_ids:
                self.highest_handled_submission_id = max(kid_ids)
            await asyncio.sleep(POLL_INTERVAL)

    async def get_user_submissions(self) -> list[int]:
        log.debug("Getting user subs: %s", self.hn_username)
        async with self.http_session.get(
            f"{API_URL}/user/{self.hn_username}.json"
        ) as r:
            if r.status != 200:
                log.warning("Bad response from API: %s", r)
                raise RuntimeError
            return (await r.json())["submitted"]

    async def send_own_and_reply(self, user_submission, reply_id):
        contact: LegacyContact = self.contacts.by_legacy_id(reply_id)
        date = datetime.fromtimestamp(user_submission["time"])
        contact.carbon(
            parse_comment_text(user_submission["text"]),
            date=date,
        )
        kid = await self.get_item(reply_id)
        contact.send_text(parse_comment_text(kid["text"]))

    async def get_item(self, item_id):
        async with self.http_session.get(f"{API_URL}/item/{item_id}.json") as r:
            return await r.json()

    async def logout(self, p: Optional[Presence]):
        pass

    async def send_text(self, t: str, c: LegacyContact):
        goto = f"threads?id={self.hn_username}#{c.legacy_id}"
        url = f"{REPLY_URL}?id={c.legacy_id}&goto={goto}"
        async with self.http_session.get(url) as r:
            if r.status != 200:
                raise XMPPError(text="Couldn't get the post reply web page from HN")
            form_page_content = await r.text()
        match = re.search(HMAC_RE, form_page_content)

        if match is None:
            raise XMPPError(
                text="Couldn't find the HMAC hidden input on the comment reply thread"
            )

        await asyncio.sleep(SLEEP_BEFORE_POST)

        form_dict = {
            "hmac": match.group(1),
            "parent": c.legacy_id,
            "text": t,
            "goto": goto,
        }

        form = aiohttp.FormData(form_dict)
        async with self.http_session.post(REPLY_POST_URL, data=form) as r:
            first_attempt_html_content = await r.text()

        log.debug("Reply response #1: %s", r)
        if r.status != 200:
            raise XMPPError(text=f"Problem replying: {r}")

        if REPOST_TEXT in first_attempt_html_content:
            match = re.search(HMAC_RE, first_attempt_html_content)
            if match is None:
                raise XMPPError(
                    text="We should repost but haven't found any hmac field on the repost page"
                )

            await asyncio.sleep(SLEEP_BEFORE_POST2)
            form = aiohttp.FormData(form_dict | {"hmac": match.group(1)})
            async with self.http_session.post(REPLY_POST_URL, data=form) as r:
                log.debug("Reply response #2: %s", r)
                if r.status != 200:
                    raise XMPPError(text=f"Problem replying: {r}")

    # none of the following make sense in a HN context,
    # this is just to avoid raising NotImplementedErrors
    async def send_file(self, u: str, c: LegacyContact):
        pass

    async def active(self, c: LegacyContact):
        pass

    async def inactive(self, c: LegacyContact):
        pass

    async def composing(self, c: LegacyContact):
        pass

    async def paused(self, c: LegacyContact):
        pass

    async def displayed(self, legacy_msg_id: Any, c: LegacyContact):
        pass

    async def correct(self, text: str, legacy_msg_id: Any, c: LegacyContact):
        pass

    async def search(self, form_values: Dict[str, str]):
        pass


def parse_comment_text(text: str):
    # TODO: use regex or something more efficient here
    return text.replace("<p>", "\n").replace("&#x27;", "'").replace("&#x2F;", "/")


LOGIN_URL = "https://news.ycombinator.com/login"
REPLY_URL = "https://news.ycombinator.com/reply"
REPLY_POST_URL = "https://news.ycombinator.com/comment"
API_URL = "https://hacker-news.firebaseio.com/v0"
POLL_INTERVAL = 30  # seconds
SLEEP_BEFORE_POST = 30  # seconds
SLEEP_BEFORE_POST2 = 5  # seconds
REPOST_TEXT = "<tr><td>Please confirm that this is your comment by submitting it one"
HMAC_RE = re.compile(r'name="hmac" value="([a-zA-Z\d]*)"')

log = logging.getLogger(__name__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slidge',
 'slidge.legacy',
 'slidge.plugins',
 'slidge.plugins.signal',
 'slidge.plugins.telegram',
 'slidge.xep_0055',
 'slidge.xep_0077',
 'slidge.xep_0100',
 'slidge.xep_0115',
 'slidge.xep_0333',
 'slidge.xep_0356',
 'slidge.xep_0363']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.5.3,<2.0.0',
 'Pillow>=8.1.0',
 'aiohttp>=3.8.1',
 'qrcode>=7.3',
 'slixmpp>=1.8.2,<2.0.0']

extras_require = \
{'discord': ['nextcord>=2.0.0-alpha.10,<3.0.0'],
 'facebook': ['mautrix-facebook>=0.4.0,<0.5.0'],
 'mattermost': ['mattermostdriver>=7.3.2,<8.0.0'],
 'signal': ['aiosignald>=0.2.1,<0.3.0'],
 'skype': ['SkPy>=0.10.4,<0.11.0'],
 'telegram': ['aiotdlib>=0.19.2,<0.20.0', 'pydantic']}

entry_points = \
{'console_scripts': ['slidge = slidge.__main__:main']}

setup_kwargs = {
    'name': 'slidge',
    'version': '0.1.0a1',
    'description': 'XMPP bridging framework',
    'long_description': "Slidge 🛷\n========\n\nPythonic XMPP gateways.\n\n[![Documentation status](https://readthedocs.org/projects/slidge/badge/?version=latest)](https://slidge.readthedocs.io/)\n[![builds.sr.ht status](https://builds.sr.ht/~nicoco/slidge/commits/master/.build.yml.svg)](https://builds.sr.ht/~nicoco/slidge/commits/master/.build.yml?)\n[![pypi](https://badge.fury.io/py/slidge.svg)](https://pypi.org/project/slidge/)\n\nSlidge is a general purpose XMPP gateway framework using the python\n\nHomepage: [sourcehut](https://sr.hr/~nicoco/slidge)\n\nChat room:\n[slidge\\@conference.nicoco.fr](xmpp:slidge@conference.nicoco.fr?join)\n\nIssue tracker: https://todo.sr.ht/~nicoco/slidge\n\nStatus\n------\n\nSlidge is alpha-grade software!\nRight now, only direct messages are implemented, no group chat stuff at all.\nDirect messaging does (more or less) work for the 5 plugins included in this repo though:\nTelegram, Signal, Facebook messenger, Skype and Hackernews.\n\nTesting locally should be fairly easy, so please go ahead and give me some\nfeedback, through the [MUC](xmpp:slidge@conference.nicoco.fr?join), the\n[issue tracker](https://todo.sr.ht/~nicoco/slidge) or in my\n[public inbox](https://lists.sr.ht/~nicoco/public-inbox).\n\nInstallation\n------------\n\nThe easiest way to try out slidge is with docker-compose. Clone the\nrepo, run `docker-compose up` and you should have:\n\n-   an XMPP server (prosody) exposed on port 5222 with a registered user\n    <test@localhost> (password: password)\n-   3 gateway components (a dummy network, signal and telegram)\n-   hot reloading of gateways on code change\n-   signald running in a container (required for signal)\n\nI recommend using gajim to test it. You can launch it with the -p option\nto use a clean profile and not mess up your normal user settings and\nsuch.\n\nIt is definitely possible to set up everything without docker, but note\nthat the aiotdlib package needs to be manually built (wheels from pypi\nare incomplete unfortunately).\n\nAbout privacy\n-------------\n\nSlidge (and most if not all XMPP gateway that I know of) will break\nend-to-end encryption, or more precisely one of the \\'ends\\' become the\ngateway itself. If privacy is a major concern for you, my advice would\nbe to:\n\n-   use XMPP + OMEMO\n-   self-host your gateways\n-   have your gateways hosted by someone you know AFK\n\nRelated projects\n----------------\n\n-   [Spectrum](https://www.spectrum.im/)\n-   [Bitfrost](https://github.com/matrix-org/matrix-bifrost)\n-   [Mautrix](https://github.com/mautrix)\n-   [matterbridge](https://github.com/42wim/matterbridge)\n-   [XMPP-discord-bridge](https://git.polynom.me/PapaTutuWawa/xmpp-discord-bridge)\n\n",
    'author': 'Nicolas Cedilnik',
    'author_email': 'nicoco@nicoco.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://sr.ht/~nicoco/slidge/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
